! WHIZARD 2.2.0_alpha-3 Oct 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_circe1

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use limits, only: CIRCE1_EPSILON !NODEP!
  use diagnostics !NODEP!
  use tao_random_numbers !NODEP!
  use lorentz !NODEP!
  use models
  use flavors
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base
  use circe1 !NODEP!

  implicit none
  private

  public :: circe1_data_t 

  integer, parameter :: NONE = 0
  integer, parameter :: FLV_INAPPLICABLE = 1

  type, extends (sf_data_t) :: circe1_data_t 
     private 
     type(model_t), pointer :: model => null () 
     type(flavor_t), dimension(2) :: flv_in 
     integer, dimension(2) :: pdg_in
     real(default), dimension(2) :: m_in = 0
     logical, dimension(2) :: photon = .false.
     logical :: generate = .true.
     type(tao_random_state), pointer :: rng => null ()
     logical :: map = .true.
     real(default), dimension(2) :: beta = 0
     real(default), dimension(2) :: gamma = 0   
     real(default) :: sqrts = 0
     integer :: ver = 0 
     integer :: rev = 0      
     integer :: acc = 0      
     integer :: chat = 0    
     integer :: error = NONE 
   contains  
       procedure :: init => circe1_data_init 
       procedure :: write => circe1_data_write 
       procedure :: get_n_par => circe1_data_get_n_par
       procedure :: get_pdg_out => circe1_data_get_pdg_out
       procedure :: allocate_sf_int => circe1_data_allocate_sf_int     
  end type circe1_data_t 
 
  !!! JRR: WK please check
  type, extends (sf_int_t) :: circe1_t
     type(circe1_data_t), pointer :: data => null ()
     real(default), dimension(2) :: x = 0
     real(default), dimension(2) :: q = 0
   contains
     procedure :: type_string => circe1_type_string
     procedure :: write => circe1_write
     procedure :: complete_kinematics => circe1_complete_kinematics
     procedure :: inverse_kinematics => circe1_inverse_kinematics
     procedure :: init => circe1_init
     procedure :: apply => circe1_apply
  end type circe1_t 
  



contains

  subroutine circe1_data_init &
       (data, model, flv, sqrts, out_photon, generate, rng, map, &
        ver, rev, acc, chat)
    class(circe1_data_t), intent(out) :: data 
    type(model_t), intent(in), target :: model
    type(flavor_t), dimension(2), intent(in) :: flv
    real(default), intent(in) :: sqrts 
    logical, dimension(2), intent(in) :: out_photon
    logical, intent(in) :: generate, map
    type(tao_random_state), intent(in), target :: rng
    integer, intent(in) :: ver, rev, acc, chat 
    data%model => model 
    data%flv_in = flv 
    data%pdg_in = flavor_get_pdg (data%flv_in)
    if (any (abs (data%pdg_in) /= ELECTRON))  data%error = FLV_INAPPLICABLE
    if (data%pdg_in(1) /= - data%pdg_in(2))  data%error = FLV_INAPPLICABLE
    data%m_in = flavor_get_mass (data%flv_in)
    data%sqrts = sqrts
    data%photon = out_photon
    data%generate = generate
    data%rng => rng
    data%map = map
    data%ver = ver
    data%rev = rev
    data%acc = acc 
    data%chat = chat 
    select case (char (model_get_name (data%model)))
    case ("QCD","Test")
       call msg_fatal ("CIRCE1 structure function not available for model " &
            // char (model_get_name (data%model)) // ".")
    end select         
    call circes (0.d0, 0.d0, dble (data%sqrts), &
         data%acc, data%ver, data%rev, data%chat)
    call circe1_get_parameters (data%photon, data%beta, data%gamma)
  end subroutine circe1_data_init 
 
  elemental subroutine circe1_get_parameters (photon, beta, gamma)
    logical, intent(in) :: photon
    real(default), intent(out) :: beta, gamma
    if (photon) then
       beta = circe1_params%a1(6)
       gamma = circe1_params%a1(5)
    else
       beta = circe1_params%a1(2)
       gamma = circe1_params%a1(3)
    end if
  end subroutine circe1_get_parameters
    
  subroutine circe1_data_write (data, unit, verbose) !, md5)
    class(circe1_data_t), intent(in) :: data 
    integer, intent(in), optional :: unit 
    logical, intent(in), optional :: verbose
    integer :: u
    !!! JRR: WK please check: still needed?
!    logical, intent(in), optional :: md5
    u = output_unit (unit);  if (u < 0)  return 
    write (u, "(1x,A)") "CIRCE1 data:" 
    write (u, "(3x,A,A,A,A)") "  prt_in  = ", &
         char (flavor_get_name (data%flv_in(1))), &
         ", ", char (flavor_get_name (data%flv_in(2)))
    write (u, "(3x,A,L1)")      "  photon  = ", data%photon
    write (u, "(3x,A,L1)")      "  generate = ", data%generate
    write (u, "(3x,A,L1)")      "  map      = ", data%map
    write (u, "(1x,A,ES19.12)") "  beta  = ", data%beta 
    write (u, "(1x,A,ES19.12)") "  gamma = ", data%gamma 
    write (u, "(1x,A,ES19.12)") "  m_in  = ", data%m_in
    write (u, "(1x,A,ES19.12)") "  sqrts = ", data%sqrts 
    write (u, "(3x,I3)")        "  ver = ", data%ver 
    write (u, "(3x,I3)")        "  rev = ", data%rev 
    write (u, "(3x,I3)")        "  acc = ", data%acc 
    write (u, "(3x,A,L1)")      "  chat = ", data%chat 
  end subroutine circe1_data_write 
 
  function circe1_data_get_n_par (data) result (n)
    class(circe1_data_t), intent(in) :: data
    integer :: n
    n = 2
  end function circe1_data_get_n_par
  
!!! JRR: WK please check
  function circe1_data_get_pdg_out (data) result (pdg_out)
    class(circe1_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    integer :: i, n
    n = 2
    allocate (pdg_out (n))
    do i = 1, n
       if (data%photon(i)) then
          pdg_out(i) = PHOTON
       else
          pdg_out(i) = data%pdg_in(i)
       end if
    end do
  end function circe1_data_get_pdg_out
  
  subroutine circe1_data_allocate_sf_int (data, sf_int)
    class(circe1_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (circe1_t :: sf_int)
  end subroutine circe1_data_allocate_sf_int
  
  function circe1_type_string (object) result (string)
    class(circe1_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "CIRCE1: beamstrahlungs spectrum" 
    else
       string = "CIRCE1: [undefined]"
    end if
  end function circe1_type_string
  
  subroutine circe1_write (object, unit)
    !!! JRR: WK please check
    !!! Guess these variables do not exist for CIRCE1 (?)
    class(circe1_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "CIRCE1 data: [undefined]"
    end if
  end subroutine circe1_write
    
  subroutine circe1_complete_kinematics (sf_int, x, f, r, map)
    !!! JRR: WK please check
    class(circe1_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default), dimension(2) :: xb1
    if (map) then
       call msg_fatal ("CIRCE1: map flag not supported")
    else
       x = r
       f = 1
    end if
    xb1 = 1 - x
    call sf_int%split_momenta (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1:2)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine circe1_complete_kinematics

  subroutine circe1_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    !!! JRR: WK please check
    !!! This cannot be correct, as the CIRCE1 structure function has 
    !!! twice the variables (2->4 instead of 1->2 splitting)
    class(circe1_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default), dimension(2) :: xb1
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (map) then
       call msg_fatal ("CIRCE1: map flag not supported")
    else
       r = x
       f = 1
    end if
    if (set_mom) then
       xb1 = 1 - x
       call sf_int%split_momenta (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1:2)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine circe1_inverse_kinematics

  subroutine circe1_init (sf_int, data) 
    class(circe1_t), intent(out) :: sf_int 
    class(sf_data_t), intent(in), target :: data
    logical, dimension(6) :: mask_h
    type(quantum_numbers_mask_t), dimension(6) :: mask 
    integer, dimension(6) :: hel_lock 
    type(polarization_t) :: pol1, pol2
    type(quantum_numbers_t), dimension(1) :: qn_fc1, qn_hel1, qn_fc2, qn_hel2
    type(flavor_t) :: flv_photon
    real(default), dimension(2) :: mi2, mr2, mo2
    type(quantum_numbers_t) :: qn_photon, qn1, qn2
    type(quantum_numbers_t), dimension(6) :: qn
    type(state_iterator_t) :: it_hel1, it_hel2
    hel_lock = 0
    mask_h = .false.
    select type (data)
    type is (circe1_data_t)
       mi2 = data%m_in**2
       if (data%photon(1)) then
          hel_lock(1) = 3;  hel_lock(3) = 1;  mask_h(5) = .true.
          mr2(1) = mi2(1)
          mo2(1) = 0._default
       else
          hel_lock(1) = 5;  hel_lock(5) = 1;  mask_h(3) = .true.
          mr2(1) = 0._default
          mo2(1) = mi2(1)
       end if
       if (data%photon(2)) then
          hel_lock(2) = 4;  hel_lock(4) = 2;  mask_h(6) = .true.
          mr2(2) = mi2(2)
          mo2(2) = 0._default
       else
          hel_lock(2) = 6;  hel_lock(6) = 2;  mask_h(4) = .true.
          mr2(2) = 0._default
          mo2(2) = mi2(2)
       end if
       mask = new_quantum_numbers_mask (.false., .false., mask_h)
       call sf_int%base_init (mask, mi2, mr2, mo2, &
            hel_lock = hel_lock)
       sf_int%data => data
       call flavor_init (flv_photon, PHOTON, data%model)
       call quantum_numbers_init (qn_photon, flv_photon)
       call polarization_init_generic (pol1, data%flv_in(1))
       call quantum_numbers_init (qn_fc1(1), flv = data%flv_in(1))
       call polarization_init_generic (pol2, data%flv_in(2))
       call quantum_numbers_init (qn_fc2(1), flv = data%flv_in(2))
       call state_iterator_init (it_hel1, pol1%state) 
       do while (state_iterator_is_valid (it_hel1)) 
          qn_hel1 = state_iterator_get_quantum_numbers (it_hel1)
          qn1 = qn_hel1(1) .merge. qn_fc1(1) 
          qn(1) = qn1
          if (data%photon(1)) then
             qn(3) = qn1;  qn(5) = qn_photon
          else
             qn(3) = qn_photon;  qn(5) = qn1
          end if
          call state_iterator_init (it_hel2, pol2%state) 
          do while (state_iterator_is_valid (it_hel2)) 
             qn_hel2 = state_iterator_get_quantum_numbers (it_hel2) 
             qn2 = qn_hel2(1) .merge. qn_fc2(1) 
             qn(2) = qn2
             if (data%photon(2)) then
                qn(4) = qn2;  qn(6) = qn_photon
             else
                qn(4) = qn_photon;  qn(6) = qn2
             end if
             call interaction_add_state (sf_int%interaction_t, qn)
             call state_iterator_advance (it_hel2) 
          end do
          call state_iterator_advance (it_hel1)
       end do
       call polarization_final (pol1)
       call polarization_final (pol2)
       call interaction_freeze (sf_int%interaction_t) 
       !!! JRR: WK please check
       !!! Is this correct if the photon is the radiating particle??
       call sf_int%set_incoming ([1,2])
       call sf_int%set_radiated ([3,4])
       call sf_int%set_outgoing ([5,6])
       sf_int%status = SF_INITIAL
    end select
  end subroutine circe1_init

  subroutine circe1_apply (sf_int, scale) !!!, no_map) 
  !!! subroutine circe1_apply (sf_int, r, no_map) 
    !!! JRR: WK please check
    !!! this somehow is different from the PDFs, so the types    
    !!! do not match. Guess we need two types of strfun,
    !!! 1->2 and 2->4, nez-pas?
    class(circe1_t), intent(inout) :: sf_int
    !!! Dummy
    real(default), intent(in) :: scale
    !!!real(default), dimension(2), intent(in) :: r 
    !!!logical, intent(in) :: no_map
    !!!type(vector4_t), dimension(2) :: k
    !!!type(splitting_data_t), dimension(2) :: sd
    !!!real(default), dimension(2) :: m_in, x, xb
    real(default) :: f
    !!!type(vector4_t), dimension(4) :: q
    associate (data => sf_int%data)
      !!! !!! JRR: WKK please check
      !!! !!! No clue what to do here :(((
      !!! k(1) = interaction_get_momentum (sf_int%interaction_t, 1) 
      !!! k(2) = interaction_get_momentum (sf_int%interaction_t, 2) 
      !!! m_in = data%m_in
      !!! sd = new_splitting_data (k, m_in**2, m_in**2, m_in) 
      !!! call strfun (f, x, xb, r, circe1_data, no_map)
      !!! call splitting_set_t_bounds (sd, x, xb) 
      !!! call splitting_set_collinear (sd)
      !!! q((/1, 3/)) = split_momentum (k(1), sd(1))
      !!! q((/2, 4/)) = split_momentum (k(2), sd(2))
      !!! call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.) 
    end associate
    call interaction_set_matrix_element &
        (sf_int%interaction_t, cmplx (f, kind=default)) 
    sf_int%status = SF_EVALUATED
  end subroutine circe1_apply
 

end module sf_circe1
