! WHIZARD 2.2.0_alpha-3 Oct 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_circe2

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use tao_random_numbers !NODEP!
  use lorentz !NODEP!
  use models
  use flavors
  use helicities
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base
  use circe2 !NODEP!

  implicit none
  private

  public :: circe2_data_t

  type, extends (sf_data_t) :: circe2_data_t
     private
     type(flavor_t), dimension(2) :: flv_in
     integer, dimension(2) :: pdg = 0
     real(default), dimension(2) :: mass = 0
     logical :: generate = .true.
     type(tao_random_state), pointer :: rng => null ()
     logical :: map = .true.
     integer, dimension(2) :: map_mode = -1
     real(default), dimension(2) :: map_power = 0
     type(string_t) :: file 
     type(string_t) :: design 
     real(default) :: sqrts = 0
     logical :: polarized = .false.
     real(default) :: lumi = 0
     real(default), dimension(-1:1,-1:1) :: lumi_hel_frac = 0
     real(default), dimension(0:4) :: lumi_hel_sum = 0
   contains
       procedure :: init => circe2_data_init
       procedure :: write => circe2_data_write
       procedure :: get_n_par => circe2_data_get_n_par
       procedure :: get_pdg_out => circe2_data_get_pdg_out
       procedure :: allocate_sf_int => circe2_data_allocate_sf_int
  end type circe2_data_t

  !!! JRR: WK please check
  type, extends (sf_int_t) :: circe2_t
     type(circe2_data_t), pointer :: data => null ()
     real(default) :: x = 0
     real(default) :: q = 0
   contains
     procedure :: type_string => circe2_type_string
     procedure :: write => circe2_write
     procedure :: complete_kinematics => circe2_complete_kinematics
     procedure :: inverse_kinematics => circe2_inverse_kinematics
     procedure :: init => circe2_init
     procedure :: apply => circe2_apply
  end type circe2_t 
  



contains

  subroutine circe2_data_init &
       (data, flv_in, generate, rng, map, file, design, sqrts, polarized)
    class(circe2_data_t), intent(out) :: data
    type(flavor_t), dimension(2), intent(in) :: flv_in
    logical, intent(in) :: generate
    type(tao_random_state), intent(in), target :: rng
    logical, intent(in) :: map
    type(string_t), intent(in) :: file, design
    real(default), intent(in) :: sqrts
    logical, intent(in) :: polarized
    integer :: error, i, h1, h2, h
    data%flv_in = flv_in
    data%pdg = flavor_get_pdg (data%flv_in)
    data%mass = flavor_get_mass (data%flv_in)
    data%generate = generate
    data%rng => rng
    data%map = map
    if (data%map) then
       do i = 1, 2
          select case (abs (data%pdg(i)))
          case (PHOTON);    data%map_mode(i) = 0;  data%map_power(i) = 3
          case (ELECTRON);  data%map_mode(i) = 1;  data%map_power(i) = 12
          case default;  call msg_fatal &
               ("CIRCE2: defined only for photon and electron beams")
          end select
       end do
    else
       data%map_mode = -1
    end if
    data%file = file
    data%design = design
    data%sqrts = sqrts
    data%polarized = polarized
    error = 1
    call cir2ld (trim (char(data%file)), trim (char(data%design)), &
            dble (data%sqrts), error)
    select case (error)
    case (-1)
       call msg_fatal ("CIRCE2: data file not found.")
    case (-2)
       call msg_fatal ("CIRCE2: beam parameters do not match data file.")
    case (-3)
       call msg_fatal ("CIRCE2: invalid format of data file.")
    case (-4)
       call msg_fatal ("CIRCE2: data file too large.")
    end select
    data%lumi = cir2lm (data%pdg(1), 0, data%pdg(2), 0)
    if (data%lumi == 0) then
       call circe2_data_write (data)
       call msg_fatal ("CIRCE2: luminosity vanishes for specified beams.")
    end if
    if (data%polarized) then
       h = 0
       do h1 = -1, 1, 2
          do h2 = -1, 1, 2
             data%lumi_hel_frac(h1,h2) = &
                  cir2lm (data%pdg(1), h1, data%pdg(2), h2) / data%lumi
             h = h + 1
             data%lumi_hel_sum(h) = &
                  data%lumi_hel_sum(h-1) + data%lumi_hel_frac(h1,h2)
          end do
       end do
       data%lumi_hel_sum(4) = 1
    end if
  end subroutine circe2_data_init

  subroutine circe2_data_write (data, unit, verbose) !, md5)
    class(circe2_data_t), intent(in) :: data
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose
    integer :: h1, h2
    integer :: u
    !!! JRR: WK please check: still needed?
!    logical, intent(in), optional :: md5
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(1x,A)") "CIRCE2 data:"
    write (u, "(3x,A,A)")       "  file = ", char(data%file)
    write (u, "(3x,A,A)")       "  design = ", char(data%design)
    write (u, "(1x,A,ES19.12)") "  sqrts = ", data%sqrts
    write (u, "(3x,A,A,A,A)")   "  prt_in  = ", &
         char (flavor_get_name (data%flv_in(1))), &
         ", ", char (flavor_get_name (data%flv_in(2)))    
    write (u, "(1x,A,ES19.12)") "  mass  = ", data%mass
    write (u, "(3x,A,L1)")      "  polarized = ", data%polarized
    write (u, "(1x,A,ES19.12)") "  luminosity = ", data%lumi
    if (data%polarized) then
       do h1 = -1, 1, 2
          do h2 = -1, 1, 2
             write (u, "(6x,'(',I2,1x,I2,')',1x,'=',1x)", advance="no") h1, h2
             write (u, "(6x, ES19.12)")  data%lumi_hel_frac(h1,h2)
          end do
       end do
    end if
    write (u, "(3x,A,L1)")      "  generate = ", data%generate
    write (u, "(3x,A,L1)")      "  map = ", data%map
    if (data%map) then
       write (u, "(3x,A,I3)")   "  mode = ", data%map_mode
       write (u, "(3x,A,I3)")   "  power = ", data%map_power
    end if
  end subroutine circe2_data_write
  function circe2_data_get_n_par (data) result (n)
    class(circe2_data_t), intent(in) :: data
    integer :: n
    n = 2
  end function circe2_data_get_n_par
  
!!! JRR: WK please check
  function circe2_data_get_pdg_out (data) result (pdg_out)
    class(circe2_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    integer :: i, n
    n = 2
    allocate (pdg_out (n))
    do i = 1, n
       pdg_out(i) = data%pdg(i)
    end do
  end function circe2_data_get_pdg_out
  
  subroutine circe2_data_allocate_sf_int (data, sf_int)
    class(circe2_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (circe2_t :: sf_int)
  end subroutine circe2_data_allocate_sf_int
  
  function circe2_type_string (object) result (string)
    !!! JRR: WK please check
    class(circe2_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "CIRCE2: " // object%data%file // ", " // object%data%design
    else
       string = "CIRCE2: [undefined]"
    end if
  end function circe2_type_string
  
  subroutine circe2_write (object, unit)
    !!! JRR: WK please check
    !!! Guess these variables do not exist for CIRCE2 (?)
    class(circe2_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "CIRCE2 data: [undefined]"
    end if
  end subroutine circe2_write
    
  subroutine circe2_complete_kinematics (sf_int, x, f, r, map)
    !!! JRR: WK please check
    !!! This cannot be correct, as the lumi linker structure function has 
    !!! twice the variables (2->4 instead of 1->2 splitting)
    class(circe2_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default) :: xb1
    if (map) then
       call msg_fatal ("CIRCE2: map flag not supported")
    else
       x(1) = r(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    call sf_int%split_momentum (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine circe2_complete_kinematics

  subroutine circe2_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    !!! JRR: WK please check
    !!! This cannot be correct, as the CIRCE2 structure function has 
    !!! twice the variables (2->4 instead of 1->2 splitting)
    class(circe2_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default) :: xb1
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (map) then
       call msg_fatal ("CIRCE2: map flag not supported")
    else
       r(1) = x(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    if (set_mom) then
       call sf_int%split_momentum (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine circe2_inverse_kinematics

  subroutine circe2_init (sf_int, data)
    !!! JRR: WK please check
    class(circe2_t), intent(out) :: sf_int
    class(sf_data_t), intent(in), target :: data
    type(polarization_t) :: pol3, pol4
    real(default), dimension(2), parameter :: pol_init = 0.5_default
    logical, dimension(4) :: mask_h
    real(default), dimension(0) :: null_array
    type(quantum_numbers_mask_t), dimension(4) :: mask
    type(quantum_numbers_t), dimension(4) :: qn_fc, qn_hel, qn
    type(state_iterator_t) :: it_hel3, it_hel4
    select type (data)
    type is (circe2_data_t)
       mask_h(1:2) = .true.
       mask_h(3:4) = .not. data%polarized       
       mask = new_quantum_numbers_mask (.false., .false., mask_h)
       call quantum_numbers_init (qn_fc(1), flv = data%flv_in(1))
       call quantum_numbers_init (qn_fc(2), flv = data%flv_in(2))
       call quantum_numbers_init (qn_fc(3), flv = data%flv_in(1))
       call quantum_numbers_init (qn_fc(4), flv = data%flv_in(2))
       if (data%polarized) then
          call polarization_init_diagonal &
               (pol3, data%flv_in(1), pol_init)
          call polarization_init_diagonal &
               (pol4, data%flv_in(1), pol_init)
       else
          call polarization_init_unpolarized (pol3, data%flv_in(1))
          call polarization_init_unpolarized (pol4, data%flv_in(1))
       end if
       call sf_int%base_init (mask, [0._default, 0._default], &
            null_array, [0._default, 0._default])    
       sf_int%data => data              
       call state_iterator_init (it_hel3, pol3%state) 
       qn(1) = qn_fc(1)
       qn(2) = qn_fc(2)
       do while (state_iterator_is_valid (it_hel3)) 
          qn_hel(3:3) = state_iterator_get_quantum_numbers (it_hel3)
          qn(3) = qn_hel(3) .merge. qn_fc(3) 
          call state_iterator_init (it_hel4, pol4%state) 
          do while (state_iterator_is_valid (it_hel4)) 
             qn_hel(4:4) = state_iterator_get_quantum_numbers (it_hel4) 
             qn(4) = qn_hel(4) .merge. qn_fc(4) 
             call interaction_add_state (sf_int%interaction_t, qn)
             call state_iterator_advance (it_hel4) 
          end do
          call state_iterator_advance (it_hel3)
       end do
       call polarization_final (pol3)
       call polarization_final (pol4)
       call interaction_freeze (sf_int%interaction_t)
       !!! JRR: WK please check      
       call sf_int%set_incoming ([1,2])
       call sf_int%set_outgoing ([3,4])
       sf_int%status = SF_INITIAL
    end select
  end subroutine circe2_init

  subroutine circe2_apply (sf_int, scale) !!!r, circe2_data, no_map)
    !!! JRR: WK please check
    !!! this somehow is different from the PDFs, so how 
    !!! do we connect things?
    class(circe2_t), intent(inout) :: sf_int
    real(default), intent(in) :: scale    
    !!! real(default), dimension(2), intent(in) :: r
    !!! type(circe2_data_t), intent(in) :: circe2_data
    !!! logical, intent(in) :: no_map
    !!! complex(default), parameter :: CZERO = 0
    !!! type(vector4_t), dimension(2) :: k
    !!! integer, dimension(2) :: h, h_gen, h_tmp
    !!! type(state_iterator_t) :: it
    !!! real(default) :: f
    !!! real(default), dimension(2) :: x
    !!! type(splitting_data_t), dimension(2) :: sd
    !!! type(vector4_t), dimension(4) :: q
    !!! k(1) = interaction_get_momentum (int, 1)
    !!! k(2) = interaction_get_momentum (int, 2)
    associate (data => sf_int%data)
    !!! if (circe2_data%generate) then
    !!!    call interaction_set_matrix_element (int, CZERO)
    !!!    if (circe2_data%polarized) then
    !!!       call circe2_select_hel &
    !!!            (h_gen, circe2_data%rng, circe2_data%lumi_hel_sum)
    !!!    else
    !!!       h_gen = 0
    !!!    end if
    !!! end if
    !!! call state_iterator_init (it, interaction_get_state_matrix_ptr (int))
    !!! LOOP_HEL: do while (state_iterator_is_valid (it))
    !!!    if (circe2_data%polarized) then
    !!!       h_tmp = helicity_get (state_iterator_get_helicity (it, 3))
    !!!       h(1) = h_tmp(1)
    !!!       h_tmp = helicity_get (state_iterator_get_helicity (it, 4))
    !!!       h(2) = h_tmp(1)
    !!!    else
    !!!       h = 0
    !!!    end if
    !!!    if (.not. circe2_data%generate .or. all (h == h_gen)) then
    !!!       call strfun (f, x, r, h, circe2_data, no_map)
    !!!       call state_iterator_set_matrix_element (it, cmplx (f, kind=default))
    !!!    end if
    !!!    call state_iterator_advance (it)
    !!! end do LOOP_HEL
    !!! sd = new_splitting_data &
    !!!      (k, circe2_data%mass**2, 0._default, circe2_data%mass)
    !!! call splitting_set_t_bounds (sd, x, 1-x)
    !!! call splitting_set_collinear (sd)
    !!! q((/1, 3/)) = split_momentum (k(1), sd(1))
    !!! q((/2, 4/)) = split_momentum (k(2), sd(2))
    !!! call interaction_set_momenta (int, q(3:4), outgoing=.true.)
    end associate
    sf_int%status = SF_EVALUATED
  end subroutine circe2_apply


end module sf_circe2
